/*
 * Copyright (C) 2007 LAAS/CNRS 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// to draw a list of triangles

// Would it be interesting to use GL_UNSIGNED_SHORT for triangle ? 
//  I think no as it is 3 * 4 + 4 = 16 / 3 * 4 + 2 without normal and coulor

// Would it be interesting to use interleaved arrays ?  I think no.
// Would it be interesting to use Vertex Buffer Objects ? I think no.
//  Rendering Huge Triangle Meshes with OpenGL 
//  http://www.sci.utah.edu/~bavoil/opengl/bavoil_trimeshes_2005.pdf

//  Interface TCL : see triangleTest.tcl
//
// We need some additional functions to access gdheTriangle and to 
//  read and save from files

#include <GL/gl.h>
#include "gdheTriangle.h"
#include <stdlib.h>
#include <stdio.h>

void gdheTriangleDraw(gdheTriangle* triangles)
{
  GLboolean vertexWasEnabled = glIsEnabled(GL_VERTEX_ARRAY);
  GLboolean normalWasEnabled = glIsEnabled(GL_NORMAL_ARRAY);
  GLboolean colorWasEnabled = glIsEnabled(GL_COLOR_ARRAY);
  // I don't know how to save actual color

  if (triangles->vertexNumber < 9) return;
  if (triangles->triangleNumber < 3) return;
  
  // we want to use a vertex array
  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(3, GL_FLOAT , 0, (GLvoid*) triangles->vertex);

  if (triangles->normalNumber == triangles->vertexNumber) {
    // we want to use a normal array
    glEnableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(GL_FLOAT, 0, (GLvoid*) triangles->normal);
  } else {
    glDisableClientState(GL_NORMAL_ARRAY);
  }
  
  if (triangles->colorNumber == triangles->vertexNumber) {
    // we want to use a color array
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(3, GL_UNSIGNED_BYTE, 0, triangles->color);
  } else {
    glDisableClientState(GL_COLOR_ARRAY);
    if (triangles->colorNumber == 3) {
      glColor3ub(triangles->color[0], 
		 triangles->color[1], triangles->color[2]);
    }
  }
  
  // we do the job
  glDrawElements(GL_TRIANGLES, 
		 triangles->triangleNumber,
		 GL_UNSIGNED_INT,
		 triangles->triangle);

  if (vertexWasEnabled) 
    glEnableClientState(GL_VERTEX_ARRAY);
  else
    glDisableClientState(GL_VERTEX_ARRAY);
  if (normalWasEnabled)
    glEnableClientState(GL_NORMAL_ARRAY);
  else
    glDisableClientState(GL_NORMAL_ARRAY);
  if (colorWasEnabled)
    glEnableClientState(GL_COLOR_ARRAY);
  else
    glDisableClientState(GL_COLOR_ARRAY);
}



int gdheTriangleArrayGLfloatSize(GLfloat** array, int* size, int newSize)
{
  if (*size == newSize) return 0; // nothing to do
  if (*size > newSize) return 0;
  *array = (GLfloat*) realloc(*array, sizeof(GLfloat) * newSize);
  if (*array == 0) {
    *size = 0;
    return -1;
  }
  *size = newSize;
  return 0;
}

int 
gdheTriangleArrayGLuintSize(GLuint** array, int* size, int newSize)
{
  if (*size == newSize) return 0; // nothing to do
  if (*size > newSize) return 0;
  *array = (GLuint*)
    realloc(*array, sizeof(GL_UNSIGNED_INT) * newSize);
  if (*array == 0) {
    *size = 0;
    return -1;
  }
  *size = newSize;
  return 0;
}

int
gdheTriangleArrayGLubyteSize(GLubyte** array, int* size, int newSize)
{
  if (*size == newSize) return 0; // nothing to do
  if (*size > newSize) return 0;
  *array = (GLubyte*)
    realloc(*array, sizeof(GLubyte) * newSize);
  if (*array == 0) {
    *size = 0;
    return -1;
  }
  *size = newSize;
  return 0; /* OK */
}


// resize pT->vertex using realloc ...
// return 0 if OK
int gdheTriangleVertexResize(gdheTriangle* pT, int nbVertex)
{
  if (pT == 0) return -1;
  return
    gdheTriangleArrayGLfloatSize(&(pT->vertex), &(pT->vertexSize), nbVertex);
}

// resize pT->normal using realloc ...
// return 0 if OK
int gdheTriangleNormalResize(gdheTriangle* pT, int nbNormal)
{
  if (pT == 0) return -1;
  return
    gdheTriangleArrayGLfloatSize(&(pT->normal), &(pT->normalSize), nbNormal);
}

// resize pT->color using realloc ...
// return 0 if OK
int gdheTriangleColorResize(gdheTriangle* pT, int nbColor)
{
  if (pT == 0) return -1;
  return
    gdheTriangleArrayGLubyteSize(&(pT->color), &(pT->colorSize),
					  nbColor);
}

//resize pT->triangle using realloc ...
// return 0 if OK
int gdheTriangleTriangleResize(gdheTriangle* pT, int nbTriangle)
{
  if (pT == 0) return -1;
  return
    gdheTriangleArrayGLuintSize(&(pT->triangle),
					 &(pT->triangleSize), nbTriangle);
}


int gdheTriangleAddVertex(gdheTriangle* pT, float x, float y, float z)
{
  float* pV;

  if (pT == 0) return -1;
  gdheTriangleVertexResize(pT, pT->vertexNumber + 3);
  pV = &(pT->vertex[pT->vertexNumber]);
  pV[0] = x;
  pV[1] = y;
  pV[2] = z;
  pT->vertexNumber += 3;
  return 0;
}

void gdheTriangleInit(gdheTriangle* pT)
{
  pT->vertexSize     = 0;
  pT->vertexNumber   = 0;
  pT->triangleSize   = 0;
  pT->triangleNumber = 0;
  pT->normalSize     = 0;
  pT->normalNumber   = 0;
  pT->colorSize	     = 0;
  pT->colorNumber    = 0;
  pT->vertex	     = 0;
  pT->triangle	     = 0;
  pT->normal	     = 0;
  pT->color          = 0;
}

void gdheTriangleFree(gdheTriangle* pT)
{
  pT->vertexSize     = 0;
  pT->vertexNumber   = 0;
  pT->triangleSize   = 0;
  pT->triangleNumber = 0;
  pT->normalSize     = 0;
  pT->normalNumber   = 0;
  pT->colorSize	     = 0;
  pT->colorNumber    = 0;
  if (pT->vertex != 0) free(pT->vertex);
  if (pT->triangle != 0) free(pT->triangle);
  if (pT->normal != 0) free(pT->normal);
  if (pT->color != 0) free(pT->color);
  pT->vertex	     = 0;
  pT->triangle	     = 0;
  pT->normal	     = 0;
  pT->color          = 0;
}

void gdheTriangleFreeVertex(gdheTriangle* pT)
{
  pT->vertexSize     = 0;
  pT->vertexNumber   = 0;
  if (pT->vertex != 0) free(pT->vertex);
  pT->vertex	     = 0;
}

void gdheTriangleFreeTriangle(gdheTriangle* pT)
{
  pT->triangleSize   = 0;
  pT->triangleNumber = 0;
  if (pT->triangle != 0) free(pT->triangle);
  pT->triangle	     = 0;
}

void gdheTriangleFreeNormal(gdheTriangle* pT)
{
  pT->normalSize     = 0;
  pT->normalNumber   = 0;
  if (pT->normal != 0) free(pT->normal);
  pT->normal	     = 0;
}

void gdheTriangleFreeColor(gdheTriangle* pT)
{
  pT->colorSize	     = 0;
  pT->colorNumber    = 0;
  if (pT->color != 0) free(pT->color);
  pT->color          = 0;
}

void gdheTrianglePrintFile(FILE* output, gdheTriangle* pT)
{
  GLfloat* pVertex;
  GLuint* pTriangle;
  GLfloat* pNormal;
  GLubyte* pColor;
  int i;

  fprintf(output, "vertexSize= %d\n", pT->vertexSize);
  fprintf(output, "vertexNumber= %d\n", pT->vertexNumber);
  fprintf(output, "triangleSize= %d\n", pT->triangleSize);
  fprintf(output, "triangleNumber= %d\n", pT->triangleNumber);
  fprintf(output, "normalSize= %d\n", pT->normalSize);
  fprintf(output, "normalNumber= %d\n", pT->normalNumber);
  fprintf(output, "colorSize= %d\n", pT->colorSize);
  fprintf(output, "colorNumber= %d\n", pT->colorNumber);
  fprintf(output, "vertex = { ");
  pVertex = pT->vertex;
  for (i = 0; i < pT->vertexNumber; i++) {
    fprintf(output, "%g ", *pVertex);
    pVertex++;
  }
  fprintf(output, "}\n");

  fprintf(output, "triangle = { ");
  pTriangle = pT->triangle;
  for (i = 0; i < pT->triangleNumber; i++) {
    fprintf(output, "%d ", *pTriangle);
    pTriangle++;
  }
  fprintf(output, "}\n");

  fprintf(output, "normal = { ");
  pNormal = pT->normal;
  for (i = 0; i < pT->normalNumber; i++) {
    fprintf(output, "%g ", *pNormal);
    pNormal++;
  }
  fprintf(output, "}\n");

  fprintf(output, "color = { ");
  pColor = pT->color;
  for (i = 0; i < pT->colorNumber; i++) {
    fprintf(output, "%d ", *pColor);
    pColor++;
  }
  fprintf(output, "}\n");
}

void gdheTrianglePrint(gdheTriangle* pT)
{
  gdheTrianglePrintFile(stdout, pT);
}
