/*
 * Copyright (C) 1996-2010 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"

#include <stdlib.h>
#include <math.h>
#include <tcl.h>
#include <tk.h>
#include <GL/glu.h>

#include "objets.h"
#include "global.h"

/***
 *** Procedures used to draw elementary objects
 ***/

/**
 ** local variables
 **/
static GLUtriangulatorObj *tobj = NULL;
static int concavePolygons = 0;

static double angleLimit(double angle);
static void drawbox(GLdouble x0, GLdouble y0, GLdouble z0, 
		    GLdouble x1, GLdouble y1, GLdouble z1, GLenum type);

/*----------------------------------------------------------------------*/

/**
 ** 3D segments
 **/

int
polylineCallBack(ClientData clientData, Tcl_Interp *interp, 
		 int argc, Tcl_Obj *const argv[])
{
    int i, n;
    double x, y, z;

    if (argc < 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "polyline n x0 y0 z0...  \"",
		      -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &n);
    
    if (argc != n*3+2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "polyline n x0 y0 z0...  \"",
		      -1));
	return TCL_ERROR;
    }
    glBegin(GL_LINE_STRIP);
    for (i = 0; i < n; i++) {
	Tcl_GetDoubleFromObj(interp, argv[3*i+2], &x);
	Tcl_GetDoubleFromObj(interp, argv[3*i+3], &y);
	Tcl_GetDoubleFromObj(interp, argv[3*i+4], &z);
	glVertex3d(x, y, z);
    }
    glEnd();

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
  
}

/*----------------------------------------------------------------------*/
/*
 *   angleLimit  -  Limit the value of an angle between -M_PI and M_PI
 */
static double 
angleLimit (angle)
     double angle;               /* Angle to be limited */
 
{
  /* adjust angles that are > M_PI */
  while (angle > M_PI)
    angle -= 2.*M_PI;
 
  /* adjust angles < -M_PI */
  while (angle < -M_PI)
    angle += 2.*M_PI;
 
  return (angle);
}

/**
 ** Reed&Shep style curves in the z=0 plane
 **/
int
reedShepCallBack(ClientData clientData, Tcl_Interp *interp,
		 int argc, Tcl_Obj *const argv[])
{
    int i, j, n;
    double x, y, theta, dt, dl;

    if (argc < 5) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # of args: should be "
		      "\"reedShep x y theta n dt dl ...\"",
		      -1));
	return TCL_ERROR;
    }
    Tcl_GetDoubleFromObj(interp, argv[1], &x);
    Tcl_GetDoubleFromObj(interp, argv[2], &y);
    Tcl_GetDoubleFromObj(interp, argv[3], &theta);
    theta = angleLimit(theta);
    Tcl_GetIntFromObj(interp, argv[4], &n);
    
    if (n == 0) 
	    return TCL_OK;
      
    if ((n < 0) || (argc != 5+n*2)) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("incoherent # of args", -1));
	return TCL_ERROR;
    }


    glBegin(GL_LINE_STRIP);
    glVertex2d(x, y);
    for (i = 0; i < n; i++) {
	Tcl_GetDoubleFromObj(interp, argv[5+2*i], &dt);
	Tcl_GetDoubleFromObj(interp, argv[5+2*i+1], &dl);
	/* next starting point */
	if (dt == 0.0) {
	    /* line segment - easy */
	    x +=  dl*cos(theta);
	    y +=  dl*sin(theta);
	    glVertex2d(x, y);
	} else {
	    /* circular arc divided into 6 segments */
	    dt = dt/6;
	    dl = dl/6;
	    for (j = 0; j < 6; j++) {
		x += dl/dt*(sin(theta+dt) - sin(theta));
		y += dl/dt*(cos(theta) - cos(theta+dt));
		glVertex2d(x, y);
		theta = angleLimit(theta + dt);
	    } /* for */
	}
    } /* for */

    glEnd();
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/* drawbox:
 *
 * draws a rectangular box with the given x, y, and z ranges.  
 * The box is axis-aligned.
 */
static void 
drawbox(GLdouble x0, GLdouble y0, GLdouble z0, 
	GLdouble x1, GLdouble y1, GLdouble z1, 
	GLenum type)
{
    static GLdouble n[6][3] = {
        {-1.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {1.0, 0.0, 0.0},
        {0.0, -1.0, 0.0}, {0.0, 0.0, 1.0}, {0.0, 0.0, -1.0}
    };
    static GLint faces[6][4] = {
        { 0, 1, 2, 3 }, { 3, 2, 6, 7 }, { 7, 6, 5, 4 },
        { 4, 5, 1, 0 }, { 5, 6, 2, 1 }, { 7, 4, 0, 3 }
    };
    GLdouble v[8][3], tmp;
    GLint i;
 
    if (x0 > x1) {
        tmp = x0; x0 = x1; x1 = tmp;
    }
    if (y0 > y1) {
        tmp = y0; y0 = y1; y1 = tmp; 
    }
    if (z0 > z1) {
        tmp = z0; z0 = z1; z1 = tmp; 
    }
    v[0][0] = v[1][0] = v[2][0] = v[3][0] = x0;
    v[4][0] = v[5][0] = v[6][0] = v[7][0] = x1;
    v[0][1] = v[1][1] = v[4][1] = v[5][1] = y0;
    v[2][1] = v[3][1] = v[6][1] = v[7][1] = y1;
    v[0][2] = v[3][2] = v[4][2] = v[7][2] = z0;
    v[1][2] = v[2][2] = v[5][2] = v[6][2] = z1;
 
    glBegin(type);
    for (i = 0; i < 6; i++) {
        glNormal3dv(&n[i][0]);
        glVertex3dv(&v[faces[i][0]][0]);
        glVertex3dv(&v[faces[i][1]][0]);
        glVertex3dv(&v[faces[i][2]][0]);
        glVertex3dv(&v[faces[i][3]][0]);
    }
    glEnd();
}

/*----------------------------------------------------------------------*/

/**
 ** box with edges parallel to the main axes
 **/
int
boxCallBack(ClientData clientData, Tcl_Interp *interp, 
	    int argc, Tcl_Obj *const argv[])
{
    double xc, yc, zc, dx, dy, dz;
    double x1, y1, z1, x2, y2, z2;
    
    if (argc != 7) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "drawCube ?x1? ?y1? ?z1? ?x2? ?y2? ?z2?\"",
		      -1));
	return TCL_ERROR;
    }
    Tcl_GetDoubleFromObj(interp, argv[1], &xc);
    Tcl_GetDoubleFromObj(interp, argv[2], &yc);
    Tcl_GetDoubleFromObj(interp, argv[3], &zc);
    Tcl_GetDoubleFromObj(interp, argv[4], &dx);
    Tcl_GetDoubleFromObj(interp, argv[5], &dy);
    Tcl_GetDoubleFromObj(interp, argv[6], &dz);
    x1 = xc - dx/2;
    y1 = yc - dy/2;
    z1 = zc;
    x2 = xc + dx/2;
    y2 = yc + dy/2;
    z2 = zc + dz;

    drawbox(x1, y1, z1, x2, y2, z2, GL_QUADS);

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/*
 * Prisms
 */


int
prismCallBack(ClientData clientData, Tcl_Interp *interp, 
	      int argc, Tcl_Obj *const argv[])
{
    double vect0[256][3];
    double vect1[256][3];
    double norm0[3], norm1[3], norm[256][3];
    double a[3], b[3], c;
    double ext [3];
    int i, n;

    if (argc <= 5) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "prism n extx exty extz x0 y0 z0 ... \"",
		      -1));
	return TCL_ERROR;
    }
    	    
    /* Get the number of vertices */
    Tcl_GetIntFromObj(interp, argv[1], &n);
    if ( n == 0 ) {
	return(TCL_ERROR);
    }
    if (argc !=  5+n*3) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "prism n extx exty extz x0 y0 z0 ... \"",
		      -1));
	return TCL_ERROR;
    }

    /* Get the vector */
    Tcl_GetDoubleFromObj(interp, argv[2], &ext[0]);
    Tcl_GetDoubleFromObj(interp, argv[3], &ext[1]);
    Tcl_GetDoubleFromObj(interp, argv[4], &ext[2]);
    /* Get the vertex coordinates */
    for (i=0; i<n; i++) {
	Tcl_GetDoubleFromObj(interp, argv[5+(3*i)], &vect0[i][0]);
	Tcl_GetDoubleFromObj(interp, argv[6+(3*i)], &vect0[i][1]);
	Tcl_GetDoubleFromObj(interp, argv[7+(3*i)], &vect0[i][2]);
    }
    /* Compute all vertices */
    a[0] = vect0[1][0] - vect0[0][0];
    a[1] = vect0[1][1] - vect0[0][1];
    a[2] = vect0[1][2] - vect0[0][2];
    b[0] = vect0[2][0] - vect0[0][0];
    b[1] = vect0[2][1] - vect0[0][1];
    b[2] = vect0[2][2] - vect0[0][2];
    norm0[0] = a[1]*b[2] - b[1]*a[2];
    norm0[1] = a[2]*b[0] - b[2]*a[0];
    norm0[2] = a[0]*b[1] - b[0]*a[1];
    c = sqrt ( norm0[0]*norm0[0] + norm0[1]*norm0[1] + norm0[2]*norm0[2] );
    norm0[0] = norm0[0]/c;
    norm0[1] = norm0[1]/c;
    norm0[2] = norm0[2]/c;

    c = norm0[0]*ext[0] + norm0[1]*ext[1] + norm0[2]*ext[2];
    if ( c < 0.0 ) {
	for ( i = 0 ; i < n ; i ++ ) {
	    vect0[i][0] += ext[0];
	    vect0[i][1] += ext[1];
	    vect0[i][2] += ext[2];
	}
	ext[0] = -ext[0];
	ext[1] = -ext[1];
	ext[2] = -ext[2];
    }
    
    for ( i = 0 ; i < n ; i ++ ) {
	vect1[i][0] = vect0[i][0] + ext[0];
	vect1[i][1] = vect0[i][1] + ext[1];
	vect1[i][2] = vect0[i][2] + ext[2];
    }
    
    for ( i = 0 ; i < n ; i ++ ) {
	a[0] = vect0[(i+1)%n][0] - vect0[i][0];
	a[1] = vect0[(i+1)%n][1] - vect0[i][1];
	a[2] = vect0[(i+1)%n][2] - vect0[i][2];
	norm[i][0] = a[1]*ext[2]-ext[1]*a[2];
	norm[i][1] = a[2]*ext[0]-ext[2]*a[0];
	norm[i][2] = a[0]*ext[1]-ext[0]*a[1];
	c = sqrt ( norm[i][0]*norm[i][0]
		   + norm[i][1]*norm[i][1]
		   + norm[i][2] * norm[i][2] );
	norm[i][0] = norm[i][0] / c;
	norm[i][1] = norm[i][1] / c;
	norm[i][2] = norm[i][2] / c;
    }
    
    norm1[0] = -norm0[0];
    norm1[1] = -norm0[1];
    norm1[2] = -norm0[2];
	    
    /*
     * Now for the object itself 
     */

    if (concavePolygons) {
	/* Define a new tesselation object */
	if (tobj == NULL) {
	    tobj = gluNewTess();
	    if (tobj == NULL) {
		fprintf(stderr, "pb newTess\n");
	    }
	    
	    gluTessCallback(tobj, GLU_BEGIN, glBegin);
	    gluTessCallback(tobj, GLU_VERTEX, glVertex3dv);
	    gluTessCallback(tobj, GLU_END, glEnd);
	}
	
	/* upper face */
	gluBeginPolygon(tobj);
	glNormal3dv(norm1);
	for (i = 0 ; i < n ; i++) {
	    gluTessVertex(tobj,vect0[n-i-1],vect0[n-i-1]);
	}
	gluEndPolygon(tobj);
	/* lower face  */
	gluBeginPolygon(tobj);
	glNormal3dv(norm0);
	for ( i = 0 ; i < n ; i ++ ) {
	    gluTessVertex(tobj,vect1[i], vect1[i]);
	}
	gluEndPolygon(tobj);
    } else {
	/* convex polygons */

	/* upper face */
	glBegin(GL_POLYGON);
	glNormal3dv(norm1);
	for (i = 0; i < n; i++) {
	    glVertex3dv(vect0[n-i-1]);
	} 
	glEnd();
	/* lower face  */
	glBegin(GL_POLYGON);
	glNormal3dv(norm0);
	for (i = 0; i < n; i++) {
	    glVertex3dv(vect1[i]);
	} 
	glEnd();
    }
	      
    /* edges */
    glShadeModel(GL_FLAT);
    glBegin(GL_QUAD_STRIP);
    glNormal3dv(norm[0]);
    glVertex3dv(vect1[0]);
    glVertex3dv(vect0[0]);
    for (i = 0; i < n ; i++) {
	glVertex3dv(vect1[i]);
	glVertex3dv(vect0[i]);
	glNormal3dv(norm[i]);
    }
    glVertex3dv(vect1[0]);
    glVertex3dv(vect0[0]);
    glEnd();
    glShadeModel(GL_SMOOTH); 

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

    
/*----------------------------------------------------------------------*/
/**
 ** Cylinders and cones
 **/
int
cylinderCallBack(ClientData clientData, Tcl_Interp *interp, 
		 int argc, Tcl_Obj *const argv[])
{
    GLUquadricObj *quadObj;
    double xc, yc, zc, diam1, diam2, depth, r;
    char *arg;
    int nfacets = 12;


    if (argc != 8 && argc != 7 && argc != 9) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "cylinder xc yc zc axe diam1 ?diam2? depth ?facets?\"",
		      -1));
	return TCL_ERROR;
    }

    Tcl_GetDoubleFromObj(interp, argv[1], &xc);
    Tcl_GetDoubleFromObj(interp, argv[2], &yc);
    Tcl_GetDoubleFromObj(interp, argv[3], &zc);

    if ( argc == 7) {
	/* cylindre */
	Tcl_GetDoubleFromObj(interp, argv[5], &diam1);
	diam2 = diam1;
	Tcl_GetDoubleFromObj(interp, argv[6], &depth);
    } else {
	/* cone */
	Tcl_GetDoubleFromObj(interp, argv[5], &diam1);
	Tcl_GetDoubleFromObj(interp, argv[6], &diam2);
	Tcl_GetDoubleFromObj(interp, argv[7], &depth);
	if (argc == 9) {
	    Tcl_GetIntFromObj(interp, argv[8], &nfacets);
	}
    }

    glPushMatrix();
    
    /*
     * prepare the transformation of the unit cylinder
     */
    glTranslated(xc, yc, zc);
	
    arg = Tcl_GetStringFromObj(argv[4], NULL);
    switch (arg[0]) {
      case 'x':
	glRotated(90, 0, 1, 0);
	break;
      case 'y':
	glRotated(-90, 1, 0, 0);
	break;
      case 'z':
	break;
      default:
	glRotated(90, 0, 1, 0);
	Tcl_GetDoubleFromObj(interp, argv[4], &r);
	glRotated(r, 1, 0, 0 );
	break;

    } /* switch */

    /* edges */
    quadObj = gluNewQuadric ();
    gluQuadricDrawStyle (quadObj, GLU_FILL);
    gluQuadricNormals(quadObj, GLU_SMOOTH);
    if (depth >= 0.0) {
	gluCylinder (quadObj, fabs(diam1)/2, fabs(diam2)/2, depth, 
		     nfacets, 2);
    } else {
	glPushMatrix();
	glRotated(180, 0, 1, 0);
	gluCylinder (quadObj, fabs(diam1)/2, fabs(diam2)/2, -depth, 
		     nfacets, 2);
	glPopMatrix();
    }
	
    gluDeleteQuadric(quadObj);

    if (diam1 > 0) {
	/* lower face */
	quadObj = gluNewQuadric();
	gluQuadricDrawStyle (quadObj, GLU_FILL);
	gluQuadricNormals(quadObj, GLU_SMOOTH);
	gluQuadricOrientation(quadObj, GLU_INSIDE);
	gluDisk(quadObj, 0, fabs(diam1)/2, nfacets, 1);
	gluDeleteQuadric(quadObj);
    }
    if (diam2 > 0) {
	/* upper face  */
	glTranslated(0, 0, depth);
	quadObj = gluNewQuadric();
	gluQuadricDrawStyle (quadObj, GLU_FILL);
	gluQuadricNormals(quadObj, GLU_SMOOTH);
	gluDisk(quadObj, 0, fabs(diam2)/2, nfacets, 1);
	gluDeleteQuadric(quadObj);
    }
    
    glPopMatrix();

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Planar polygon
 **/
int
polygonCallBack(ClientData clientData, Tcl_Interp *interp, 
		int argc, Tcl_Obj *const argv[])
{
    int i, n;
    double *v;
    int di;
    double v1[3], v2[3], norm[3], s;

    if (argc < 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "polygon n x0 y0 ...  \"",
		      -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &n);
    
    if (argc != n*2+2 && argc != n*3+2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "polygon n x0 y0 ?z0? ...  \"",
		      -1));
	return TCL_ERROR;
    }
    if (argc == n*3+2) {
	di = 3;
    } else {
	di = 2;
    }
    /* new tesselation object */
    if (concavePolygons && tobj == NULL) {
	tobj = gluNewTess();
	if (tobj == NULL) {
	    fprintf(stderr, "pb newTess\n");
	}
	
	gluTessCallback(tobj, GLU_BEGIN, glBegin);
	gluTessCallback(tobj, GLU_VERTEX, glVertex3dv);
	gluTessCallback(tobj, GLU_END, glEnd);
    }
    /* allocate a buffer for the tesselation  */
    v = malloc(3*n*sizeof(double));
    if (v == NULL) {
	fprintf(stderr, "polygon: malloc problem\n");
	return TCL_ERROR;
    }
    for (i = 0; i < n; i++) {
	Tcl_GetDoubleFromObj(interp, argv[2+(di*i)], &v[3*i]);
	Tcl_GetDoubleFromObj(interp, argv[3+(di*i)], &v[3*i+1]);
	if (di == 2) {
	    v[3*i+2] = 0.0;
	} else {
	    Tcl_GetDoubleFromObj(interp, argv[4+(di*i)], &v[3*i+2]);
	}
    }
    /* compute normals */
    /* vector product of the vectors defined by the first 3 points */
    for (i = 0; i < 3; i++) {
	    v1[i] = v[i] - v[3+i];
	    v2[i] = v[3+i] - v[6+i];
    }
    norm[0] = v1[1]*v2[2] - v1[2]*v2[1];
    norm[1] = v1[2]*v2[0] - v1[0]*v2[2];
    norm[2] = v1[0]*v2[1] - v1[1]*v2[0];
    s = sqrt(norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2]);
    norm[0] /= s;
    norm[1] /= s;
    norm[2] /= s;
	
    if (concavePolygons) {
	gluTessBeginPolygon(tobj, NULL);
	gluTessBeginContour(tobj);
	for (i = 0; i < n; i++) {
	    gluTessNormal(tobj, norm[0], norm[1], norm[2]);
	    gluTessVertex(tobj, &v[3*i], &v[3*i]);
	}
	gluTessEndContour(tobj);
	gluTessEndPolygon(tobj);
    } else {
	glBegin(GL_POLYGON);
	for (i=0; i< n; i++) {
	    glNormal3dv(norm);
	    glVertex3dv(&v[3*i]);
	}
	glEnd();
    }
    free(v);
    
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

#ifdef notyet
/**
 ** translucent polygons
 **/
int
transparentPolygonCallBack(ClientData clientData, Tcl_Interp *interp, 
			   int argc, Tcl_Obj *const argv[])
{
    int i, n;
    double *v;
    int di;
    double v1[3], v2[3], norm[3], s;

    if (argc < 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "polygon n x0 y0 ...  \"",
		      -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &n);
    
    if (argc != n*2+2 && argc != n*3+2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "polygon n x0 y0 ?z0? ...  \"",
		      -1));
	return TCL_ERROR;
    }
    if (argc == n*3+2) {
	di = 3;
    } else {
	di = 2;
    }
    /* defini un nouvel objet de triangulation */
    if (concavePolygons && tobj == NULL) {
	tobj = gluNewTess();
	if (tobj == NULL) {
	    fprintf(stderr, "pb newTess\n");
	}
	
	gluTessCallback(tobj, GLU_BEGIN, glBegin);
	gluTessCallback(tobj, GLU_VERTEX, glVertex3dv);
	gluTessCallback(tobj, GLU_END, glEnd);
    }
    /* allocation d'un tampon pour memoriser les sommets */
    /* Necessaire pour la triangulation */
    v = malloc(3*n*sizeof(double));
    if (v == NULL) {
	fprintf(stderr, "polygon: pb malloc\n");
	return TCL_ERROR;
    }
    for (i = 0; i < n; i++) {
	Tcl_GetDoubleFromObj(interp, argv[2+(di*i)], &v[3*i]);
	Tcl_GetDoubleFromObj(interp, argv[3+(di*i)], &v[3*i+1]);
	if (di == 2) {
	    v[3*i+2] = 0.0;
	} else {
	    Tcl_GetDoubleFromObj(interp, argv[4+(di*i)], &v[3*i+2]);
	}
    }
    /* Active la transparance par le canal alpha */
    glEnable(GL_BLEND); 
    glDepthMask(GL_FALSE);      
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    
    /* c'est parti */
    /* produit vectoriel des vecteurs des 3 premiers points */
    for (i = 0; i < 3; i++) {
	    v1[i] = v[i] - v[3+i];
	    v2[i] = v[3+i] - v[6+i];
    }
    norm[0] = v1[1]*v2[2] - v1[2]*v2[1];
    norm[1] = v1[2]*v2[0] - v1[0]*v2[2];
    norm[2] = v1[0]*v2[1] - v1[1]*v2[0];
    s = sqrt(norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2]);
    norm[0] /= s;
    norm[1] /= s;
    norm[2] /= s;
    glNormal3dv(norm);
	
    if (concavePolygons) {
	gluTessBeginPolygon(tobj, NULL);
	gluTessBeginContour(tobj);
	for (i = 0; i < n; i++) {
	    gluTessVertex(tobj, &v[3*i], &v[3*i]);
	}
	gluTessEndContour(tobj);
	glutessEndPolygon(tobj);
    } else {
	glBegin(GL_POLYGON);
	for (i=0; i< n; i++) {
	    glVertex3dv(&v[3*i]);
	}
	glEnd();
    }
    /* libere les sommets */
    free(v);

    /* Restaure le trace "normal" */
    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);


    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}
#endif

/*----------------------------------------------------------------------*/

/**
 ** sphere
 **/
int
sphereCallBack(ClientData clientData, Tcl_Interp *interp, 
		int argc, Tcl_Obj *const argv[])
{
    double x, y, z, r;
    int nfacets;
    GLUquadricObj *quadObj;

    if (argc != 5 && argc != 6) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be "
		      "\"sphere x y z radius ?facets?\"", -1));
	return TCL_ERROR;
    }
    Tcl_GetDoubleFromObj(interp, argv[1], &x);
    Tcl_GetDoubleFromObj(interp, argv[2], &y);
    Tcl_GetDoubleFromObj(interp, argv[3], &z);
    Tcl_GetDoubleFromObj(interp, argv[4], &r);
    if (argc == 6) {
	Tcl_GetIntFromObj(interp, argv[5], &nfacets);
    } else {
	nfacets = 12;
    }
    
    quadObj = gluNewQuadric ();
    gluQuadricDrawStyle (quadObj, GLU_FILL);
    gluQuadricNormals(quadObj, GLU_SMOOTH);

    glPushMatrix();
    glTranslated(x, y, z);
    gluSphere(quadObj, r, nfacets, nfacets);
    glPopMatrix();

    gluDeleteQuadric(quadObj);

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}


/*----------------------------------------------------------------------*/

/**
 ** ellipsod
 **/
int
ellipsoidCallBack(ClientData clientData, Tcl_Interp *interp, 
		  int argc, Tcl_Obj *const argv[])
{
    double x, y, z;
    int nfacets;
    double scaleX, scaleY, scaleZ;
    GLUquadricObj *quadObj;

    if (argc != 7 && argc != 8) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be "
		      "\"ellipsoid x y z dx dy dz ?facets?\"", -1));
	return TCL_ERROR;
    }
    Tcl_GetDoubleFromObj(interp, argv[1], &x);
    Tcl_GetDoubleFromObj(interp, argv[2], &y);
    Tcl_GetDoubleFromObj(interp, argv[3], &z);
    Tcl_GetDoubleFromObj(interp, argv[4], &scaleX);
    Tcl_GetDoubleFromObj(interp, argv[5], &scaleY);
    Tcl_GetDoubleFromObj(interp, argv[6], &scaleZ);
    if (argc >= 8) 
      Tcl_GetIntFromObj(interp, argv[7], &nfacets);
    else 
      nfacets = 12;
    
    quadObj = gluNewQuadric ();
    gluQuadricDrawStyle (quadObj, GLU_FILL);
    gluQuadricNormals(quadObj, GLU_SMOOTH);
    
    glPushMatrix();
    glTranslated(x, y, z);
    glScaled(scaleX, scaleY, scaleZ);
    gluSphere(quadObj, 1.0, nfacets, nfacets);
    glPopMatrix();
    
    gluDeleteQuadric(quadObj);
    
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}


/*----------------------------------------------------------------------*/
    
/**
 ** disks
 **/
int
diskCallBack(ClientData clientData, Tcl_Interp *interp, 
	    int argc, Tcl_Obj *const argv[])
{
    GLUquadricObj *quadObj;
    double xc, yc, zc, diam, r;
    char *arg;
    int nfacets;

    if (argc != 6 && argc != 7) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "disk xc yc zc axe diam ?facets?\"",
		      -1));
	return TCL_ERROR;
    }

    Tcl_GetDoubleFromObj(interp, argv[1], &xc);
    Tcl_GetDoubleFromObj(interp, argv[2], &yc);
    Tcl_GetDoubleFromObj(interp, argv[3], &zc);

    Tcl_GetDoubleFromObj(interp, argv[5], &diam);

    if (argc == 7) {
	Tcl_GetIntFromObj(interp, argv[6], &nfacets);
    } else {
	nfacets = 12;
    }

    glPushMatrix();
    
    /*
     * prepare the transformation of the unit cylinder
     */
    glTranslated(xc, yc, zc);
	
    arg = Tcl_GetStringFromObj(argv[4], NULL);
    switch (arg[0]) {
      case 'x':
	glRotated(90, 0, 1, 0);
	break;
      case 'y':
	glRotated(-90, 1, 0, 0);
	break;
      case 'z':
	break;
      default:
	glRotated(90, 0, 1, 0);
	Tcl_GetDoubleFromObj(interp, argv[4], &r);
	glRotated(r, 1, 0, 0 );
	break;
    } /* switch */
    
    quadObj = gluNewQuadric();
    gluQuadricDrawStyle (quadObj, GLU_FILL);
    gluQuadricNormals(quadObj, GLU_SMOOTH);
    gluQuadricOrientation(quadObj, GLU_OUTSIDE);
    gluDisk(quadObj, 0, fabs(diam)/2, nfacets, 1);
    gluDeleteQuadric(quadObj);

    glPopMatrix();

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/
    
/**
 ** cirle
 **/
int
circleCallBack(ClientData clientData, Tcl_Interp *interp, 
	    int argc, Tcl_Obj *const argv[])
{
    GLUquadricObj *quadObj;
    double xc, yc, zc, diam, r;
    char *arg;
    int nfacets;

    if (argc != 6 && argc != 7) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "circle xc yc zc axe diam ?facets?\"",
		      -1));
	return TCL_ERROR;
    }

    Tcl_GetDoubleFromObj(interp, argv[1], &xc);
    Tcl_GetDoubleFromObj(interp, argv[2], &yc);
    Tcl_GetDoubleFromObj(interp, argv[3], &zc);

    Tcl_GetDoubleFromObj(interp, argv[5], &diam);

    if (argc == 7) {
	Tcl_GetIntFromObj(interp, argv[6], &nfacets);
    } else {
	nfacets = 12;
    }

    glPushMatrix();
    
    /*
     * prepare the transformation of the unit cylinder
     */
    glTranslated(xc, yc, zc);
	
    arg = Tcl_GetStringFromObj(argv[4], NULL);
    switch (arg[0]) {
      case 'x':
	glRotated(90, 0, 1, 0);
	break;
      case 'y':
	glRotated(-90, 1, 0, 0);
	break;
      case 'z':
	break;
      default:
	glRotated(90, 0, 1, 0);
	Tcl_GetDoubleFromObj(interp, argv[4], &r);
	glRotated(r, 1, 0, 0 );
	break;
    } /* switch */
    
    quadObj = gluNewQuadric();
    gluQuadricDrawStyle (quadObj, GLU_SILHOUETTE);
    gluQuadricNormals(quadObj, GLU_SMOOTH);
    gluQuadricOrientation(quadObj, GLU_OUTSIDE);
    gluDisk(quadObj, 0, fabs(diam)/2, nfacets, 1);
    gluDeleteQuadric(quadObj);

    glPopMatrix();

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/** 
 ** frame
 **/
int
repereCallBack(ClientData clientData, Tcl_Interp *interp, 
		int argc, Tcl_Obj *const argv[])
{
    double length = 1.0;
    
    if (argc > 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
		      "repere ?length?\"",
		      -1));
	return TCL_ERROR;
    }
    if (argc == 2) {
	Tcl_GetDoubleFromObj(interp, argv[1], &length);
    } 
    glBegin(GL_LINES);
    glVertex3f(0, 0, 0);
    glVertex3f(length, 0, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(0, length, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(0, 0, length);
    glEnd();

    glListBase( FontBase );

    glRasterPos3f( length, 0, 0 );
    glCallLists(1, GL_UNSIGNED_BYTE, "x");
    glRasterPos3f( 0, length, 0 );
    glCallLists(1, GL_UNSIGNED_BYTE, "y");
    glRasterPos3f( 0, 0, length );
    glCallLists(1, GL_UNSIGNED_BYTE, "z");

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** images (as textures)
 **/
int
pictureCallBack(ClientData clientData, Tcl_Interp *interp, 
	      int argc, Tcl_Obj *const argv[])
{
    static GLubyte polycolor[4] = {255, 255, 255, 255};
    Tk_PhotoHandle handle;
    Tk_PhotoImageBlock img;
    double x0, y0, x1, y1, x2, y2, x3, y3;
    
    if (argc != 2 && argc != 4 && argc != 10) {
	    Tcl_SetObjResult(interp, 
		Tcl_NewStringObj("wrong # args: should be \""
		    "picture name ?width? ?height?\""
		    " or \"picture name x0 y0 x1 y1 x2 y2 x3 y3\"",
		    -1));
	    return TCL_ERROR;
    }
    if (argc == 4) {
	    Tcl_GetDoubleFromObj(interp, argv[2], &x1);
	    Tcl_GetDoubleFromObj(interp, argv[3], &y2);
	    x0 = 0.0;
	    y0 = 0.0;       
	    y1 = 0.0;
	    x2 = x1;
	    x3 = 0.0;
	    y3 = y2;
    }else if (argc == 10) {
	    Tcl_GetDoubleFromObj(interp, argv[2], &x0);	
	    Tcl_GetDoubleFromObj(interp, argv[3], &y0);	
	    Tcl_GetDoubleFromObj(interp, argv[4], &x1);	
	    Tcl_GetDoubleFromObj(interp, argv[5], &y1);	
	    Tcl_GetDoubleFromObj(interp, argv[6], &x2);	
	    Tcl_GetDoubleFromObj(interp, argv[7], &y2);	
	    Tcl_GetDoubleFromObj(interp, argv[8], &x3);	
	    Tcl_GetDoubleFromObj(interp, argv[9], &y3);	
    } else {
	    x0 = 0.0; y0 = 0.0;
	    x1 = 1.0; y1 = 0.0;
	    x2 = 1.0; y2 = 1.0;
	    x3 = 0.0; y3 = 1.0;
    }
    
    handle = Tk_FindPhoto(interp, Tcl_GetStringFromObj(argv[1], NULL));
    if (handle == NULL) {
	    Tcl_AppendResult(interp, "image \"", 
		Tcl_GetStringFromObj(argv[1], NULL), 
		"\" doesn't exist or is not a photo image", (char *)NULL);
	    return TCL_ERROR;
    }
    Tk_PhotoGetImage(handle, &img);

    if (img.pixelSize != 3 && img.pixelSize != 4) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("unsupported pixel size",
						  -1));
	return TCL_ERROR;
    }

    glPushAttrib(GL_ALL_ATTRIB_BITS);
#if 0
    glEnable(GL_DEPTH_TEST);    /* Enable depth buffering */
#endif
    glEnable(GL_TEXTURE_2D);
#if 0
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glEnable(GL_BLEND);
#endif
    glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    gluBuild2DMipmaps( GL_TEXTURE_2D, img.pixelSize, img.width, img.height,
		       img.pixelSize==3 ? GL_RGB : GL_RGBA,
		       GL_UNSIGNED_BYTE, img.pixelPtr );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );

    glColor4ubv( polycolor );
    glBegin( GL_POLYGON );
    glTexCoord2f( 0.0, 1.0 );           glVertex2f( x0, y0 );
    glTexCoord2f( 1.0, 1.0 );           glVertex2f(  x1, y1 );
    glTexCoord2f( 1.0, 0.0 );           glVertex2f(  x2,  y2 );
    glTexCoord2f( 0.0, 0.0 );           glVertex2f( x3,  y3 );
    glEnd();
    
#if 0
    glDisable( GL_BLEND );
#endif
    glPopAttrib();
    
    
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

int
concaveCallBack(ClientData clientData, Tcl_Interp *interp,
		int argc, Tcl_Obj *const argv[])
{
    if (argc != 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be "
		      "\"concave ?bool?\"", -1));
	return TCL_ERROR;
    }
    if (Tcl_GetBooleanFromObj(interp, argv[1], &concavePolygons) != TCL_OK) {
	return TCL_ERROR;
    }
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/
